from typing import Callable, List

from promptulate.agents import BaseAgent
from promptulate.agents.web_agent.prompt import SYSTEM_PROMPT_TEMPLATE
from promptulate.hook import Hook, HookTable
from promptulate.llms import BaseLLM, ChatOpenAI, ErnieBot
from promptulate.tools import DuckDuckGoTool
from promptulate.utils.logger import logger


def _build_system_prompt(prompt) -> str:
    """Build the system prompt."""
    return SYSTEM_PROMPT_TEMPLATE.format(prompt=prompt)


class WebAgent(BaseAgent):
    def __init__(
        self,
        llm: BaseLLM = None,
        hooks: List[Callable] = None,
        *args,
        **kwargs,
    ):
        super().__init__(hooks, *args, **kwargs)
        self.llm: BaseLLM = llm or ChatOpenAI(
            model="gpt-3.5-turbo-16k",
            temperature=0.0,
            enable_default_system_prompt=False,
        )
        self.stop_sequences: List[str] = ["Observation"]
        self.websearch = DuckDuckGoTool()
        self.conversation_prompt: str = ""

    def get_llm(self) -> BaseLLM:
        return self.llm

    def _run(self, prompt: str, *args, **kwargs) -> str:
        # ErnieBot built-in network search
        if self.llm.llm_type == "ErnieBot":
            self.llm = ErnieBot(
                model=self.llm.model, temperature=0.1, disable_search=False
            )
            return self.llm(prompt)

        self.conversation_prompt = _build_system_prompt(prompt)
        iterations = 0

        # Loop search until find the answer
        while True:
            llm_output: str = self.llm(
                self.conversation_prompt, stop=self.stop_sequences
            )
            logger.info(
                f"[pne] tool agent <{iterations}> current prompt: {self.conversation_prompt}"  # noqa
            )

            if "Final Answer" in llm_output:
                return llm_output.split("Final Answer:")[-1]

            self.conversation_prompt += llm_output

            # get keywords and query by websearch
            query_words: str = self._find_query_words(llm_output)
            Hook.call_hook(
                HookTable.ON_AGENT_ACTION,
                self,
                action="websearch",
                action_input=query_words,
                thought=llm_output.split("\n")[0],
            )

            query_result: str = self.websearch.run(query_words)
            Hook.call_hook(
                HookTable.ON_AGENT_OBSERVATION, self, observation=query_result
            )

            self.conversation_prompt += f"Observation: {query_result}\nThought: "
            iterations += 1

    def _find_query_words(self, answer: str) -> str:
        return answer.split("Query:")[-1].replace('"', "")



from typing import Dict, List, Union

from pydantic import Field

from promptulate.tools.base import BaseTool
from promptulate.tools.semantic_scholar.api_wrapper import SemanticScholarAPIWrapper
from promptulate.utils.core_utils import listdict_to_string


class SemanticScholarQueryTool(BaseTool):
    name: str = "semantic-scholar-query"
    description: str = (
        "A query tool around Semantic Scholar."
        "Useful for when you need to answer questions about Physics, Mathematics, "
        "Computer Science, Quantitative Biology, Quantitative Finance, Statistics, "
        "Electrical Engineering, and Economics "
        "from scientific articles on semantic scholar."
        "Input should be a search query."
    )
    api_wrapper: SemanticScholarAPIWrapper = Field(
        default_factory=SemanticScholarAPIWrapper
    )

    def _run(self, query: str, **kwargs) -> Union[str, List[Dict]]:
        """A semantic scholar paper query tool and return relevant paper query
        result."""
        result = self.api_wrapper.get_paper(query, **kwargs)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return result
        return listdict_to_string(result, item_suffix="\n\n")


class SemanticScholarReferenceTool(BaseTool):
    name: str = "semantic-scholar-reference-tool"
    description: str = ""
    api_wrapper: SemanticScholarAPIWrapper = Field(
        default_factory=SemanticScholarAPIWrapper
    )

    def _run(self, query: str, **kwargs) -> Union[str, List[Dict]]:
        result = self.api_wrapper.get_references(query, **kwargs)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return result
        return listdict_to_string(result, item_suffix="\n\n")


class SemanticScholarCitationTool(BaseTool):
    name: str = "semantic-scholar-citation-tool"
    description: str = ""
    api_wrapper: SemanticScholarAPIWrapper = Field(
        default_factory=SemanticScholarAPIWrapper
    )

    def _run(self, query: str, **kwargs) -> Union[str, List[Dict]]:
        result = self.api_wrapper.get_citations(query)
        if "return_type" in kwargs and kwargs["return_type"] == "original":
            return result
        return listdict_to_string(result, item_suffix="\n\n")



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import argparse
import os
from typing import Optional

import click
import questionary

from promptulate.agents import ToolAgent
from promptulate.config import Config
from promptulate.llms import BaseLLM, ChatOpenAI, ErnieBot
from promptulate.schema import LLMType, MessageSet
from promptulate.tools import (
    ArxivQueryTool,
    Calculator,
    DuckDuckGoTool,
    HumanFeedBackTool,
    PythonREPLTool,
    SleepTool,
)
from promptulate.tools.shell import ShellTool
from promptulate.utils.color_print import print_text
from promptulate.utils.proxy import set_proxy_mode

CFG = Config()
TOOL_MAPPING = {
    "Calculator": Calculator,
    "WebSearch": DuckDuckGoTool,
    "Python Script Executor": PythonREPLTool,
    "Arxiv Query": ArxivQueryTool,
    "Sleep": SleepTool,
    "Shell Executor": ShellTool,
    "HumanFeedBackTool": HumanFeedBackTool,
}


def get_user_input() -> Optional[str]:
    marker = (
        "# You input are here (please delete this line)\n"
        "You should save it and close the notebook after writing the prompt. (please delete this line)\n"  # noqa
        "Reply 'exit' to exit the chat.\n"
    )
    message = click.edit(marker)

    if message == "exit":
        exit()

    return message


def get_user_openai_api_key():
    import os

    api_key = questionary.password("Please enter your OpenAI API Key: ").ask()
    os.environ["OPENAI_API_KEY"] = api_key


def simple_chat(llm: BaseLLM):
    messages = MessageSet(messages=[])

    while True:
        print_text("[User] ", "blue")
        prompt = get_user_input()

        if not prompt:
            ValueError("Your prompt is None, please input something.")

        print_text(prompt, "blue")
        messages.add_user_message(prompt)

        answer = llm.predict(messages)
        messages.add_message(answer)

        print_text(f"[output] {answer.content}", "green")


def web_chat(llm: BaseLLM, model: str):
    if llm.llm_type == LLMType.ErnieBot:
        llm = ErnieBot(model=model, temperature=0.1, disable_search=False)
        while True:
            print_text("[User] ", "blue")
            prompt = get_user_input()
            if not prompt:
                ValueError("Your prompt is None, please input something.")
            print_text(prompt, "blue")
            ret = llm(prompt)
            print_text(f"[agent] {ret}", "green")

    else:
        agent = ToolAgent(tools=[DuckDuckGoTool()], llm=llm)
        while True:
            print_text("[User] ", "blue")
            prompt = get_user_input()
            if not prompt:
                ValueError("Your prompt is None, please input something.")
            print_text(prompt, "blue")
            ret = agent.run(prompt)
            print_text(f"[agent] {ret}", "green")


def agent_chat(agent: ToolAgent):
    while True:
        print_text("[User] ", "blue")
        prompt = get_user_input()
        if not prompt:
            ValueError("Your prompt is None, please input something.")
        print_text(prompt, "blue")
        ret = agent.run(prompt)
        print_text(f"[agent] {ret}", "green")


def chat():
    # get parameters
    parser = argparse.ArgumentParser(
        description="Welcome to Promptulate Chat - The best chat terminal ever!"
    )
    parser.add_argument(
        "--openai_api_key",
        help="when you first run, you should enter your openai api key",
    )
    parser.add_argument(
        "--proxy_mode", help="select openai proxy and provide [off, promptulate]"
    )
    args = parser.parse_args()

    if args.openai_api_key:
        os.environ["OPENAI_API_KEY"] = args.openai_api_key
    if args.proxy_mode:
        set_proxy_mode(args.proxy_mode)

    print_text("Hi there, here is promptulate chat terminal.", "pink")

    terminal_mode = questionary.select(
        "Choose a chat terminal:",
        choices=["Simple Chat", "Agent Chat", "Web Agent Chat", "exit"],
    ).ask()

    if terminal_mode == "exit":
        exit(0)

    model = questionary.select(
        "Choose a llm model:",
        choices=[
            "gpt-3.5-turbo",
            "gpt-4",
            "gpt-3.5-turbo-16k",
            "ernie-bot-turbo",
            "ernie-bot",
            "ernie-bot-4",
        ],
    ).ask()

    # init model
    if "gpt" in model:
        llm = ChatOpenAI(model=model, temperature=0.0)
    elif "ernie" in model:
        llm = ErnieBot(model=model, temperature=0.1)

    if terminal_mode == "Simple Chat":
        simple_chat(llm)
    elif terminal_mode == "Agent Chat":
        str_tools = questionary.checkbox(
            "Choose tools you want:", choices=list(TOOL_MAPPING.keys())
        ).ask()
        tools = []
        for str_tool in str_tools:
            tools.append(TOOL_MAPPING[str_tool]())
        agent = ToolAgent(tools=tools, llm=llm)
        agent_chat(agent)
    elif terminal_mode == "Web Agent Chat":
        web_chat(llm, model)


def main():
    chat()


if __name__ == "__main__":
    main()



from importlib.metadata import version

from promptulate.utils.color_print import print_text


def main():
    print_text(
        "🌟 Welcome to Promptulate! Let's create something amazing together!😀", "green"
    )
    print_text(f"Version: {version('promptulate')}", "blue")

    prompt = """Explore the code and contribute on GitHub: 🔗 https://github.com/Undertone0809/promptulate
Access the official documentation: 🔗 https://undertone0809.github.io/promptulate/#/
Need help or have a question? Open an issue: 🔗 https://github.com/Undertone0809/promptulate/issues
Join the development and contribute: 🔗 https://undertone0809.github.io/promptulate/#/other/contribution
    """  # noqa: E501
    print_text(prompt, "blue")


if __name__ == "__main__":
    main()



from promptulate.tools.base import Tool


class HuggingFaceTool(Tool):
    """Simple wrapper for huggingface transformers tools"""

    def __init__(self, tool, name: str, description: str, **kwargs):
        from transformers.tools import Tool as _HuggingFaceTool

        self.tool = tool
        self.name = name
        self.description = description

        if isinstance(tool, _HuggingFaceTool):
            raise ValueError("HuggingFaceTool should be type of transformers.tools")

        super().__init__(**kwargs)

    def _run(self, *args, **kwargs):
        return self.tool(*args, **kwargs)



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import logging
from typing import Dict, List, Optional

from cushy_storage.orm import BaseORMModel, CushyOrmCache
from pydantic import BaseModel, Field

from promptulate.utils.core_utils import get_cache
from promptulate.utils.singleton import singleton

logger = logging.getLogger(__name__)

KEY_MODELS_MAPPER = {
    "text-davinci-003": "gpt-3.5-turbo",
    "gpt-3.5-turbo": "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k": "gpt-3.5-turbo",
    "gpt-3.5-turbo-0301": "gpt-3.5-turbo",
    "gpt-3.5-turbo-0613": "gpt-3.5-turbo",
    "gpt-4": "gpt-4",
    "gpt-4-0314": "gpt-4",
    "gpt-4-0613": "gpt-4",
    "gpt-4-32k": "gpt-4",
    "gpt-4-32k-0314": "gpt-4",
    "gpt-4-32k-0613": "gpt-4",
}


class OpenAIKey(BaseORMModel):
    def __init__(self, model: str, key: str):
        super().__init__()
        self.model = model
        self.key = key


def _parse_openai_keys(keys: List[Dict[str, str]]) -> List[OpenAIKey]:
    """parse list of openai keys to OpenAIKey

    Args:
        keys: There are 2 kind of method to input keys
        1.
            keys = [
                {"model": "gpt-3.5-turbo","keys": "key1,key2,key3"},
                {"model": "gpt-4.0","keys": "key4,key5,key6"},
            ]
        2.
            keys = [
            {"model": "gpt-3.5-turbo","key": "key1"},
            {"model": "gpt-3.5-turbo","key": "key2"},
            {"model": "gpt-4.0","key": "key3"},
        ]
    Returns:
        List of OpenAIKey
    """
    openai_keys: List[OpenAIKey] = []
    for key in keys:
        if "key" in key and key["key"]:
            openai_keys.append(OpenAIKey(KEY_MODELS_MAPPER[key["model"]], key["key"]))
        elif "keys" in key and key["keys"]:
            _keys = key["keys"].split(",")
            for _key in _keys:
                openai_keys.append(OpenAIKey(KEY_MODELS_MAPPER[key["model"]], _key))
        else:
            raise ValueError("Key type error, your field name must be `key` or `keys`")
    return openai_keys


@singleton()
class OpenAIKeyPool(BaseModel):
    """todo provide key expiration check and token balance check"""

    cache: CushyOrmCache = Field(default_factory=get_cache)

    class Config:
        arbitrary_types_allowed = True

    def get(self, model: str) -> Optional[str]:
        """Obtain the openai key from the queue header and return it, then place the key
        at the end of queue. Use gpt4 key if the gpt3.5 key does not exist.

        Args:
            model: openai model, the option is as follows:
                "text-davinci-003","gpt-3.5-turbo","gpt-3.5-turbo-16k","gpt-3.5-turbo-0301","gpt-3.5-turbo-0613",
                "gpt-4","gpt-4-0314","gpt-4-0613","gpt-4-32k","gpt-4-32k-0314","gpt-4-32k-0613"

        Returns:
            specified openai key
        """
        model = KEY_MODELS_MAPPER[model]
        openai_key: OpenAIKey = self.cache.query(OpenAIKey).filter(model=model).first()

        if not openai_key:
            if model == "gpt-4":
                return None

            model = "gpt-4"
            openai_key: OpenAIKey = (
                self.cache.query(OpenAIKey).filter(model=model).first()
            )
            if not openai_key:
                return None

        self.cache.delete(openai_key)
        self.cache.add(openai_key)
        return openai_key.key

    def set(self, keys: List[Dict[str, str]]):
        """Set list of key to cache, you can see parameter description from
        `_parse_openai_keys()`"""
        self.cache.set(_parse_openai_keys(keys))
        logger.info("[promptulate] OpenAIKeyPool set key successfully.")

    def add(self, keys: List[Dict[str, str]]):
        """add list of keys to cache, you can see parameter description from
        `_parse_openai_keys()`"""
        self.cache.add(_parse_openai_keys(keys))

    def delete(self, key: str, model: Optional[str] = None):
        if model:
            openai_key: OpenAIKey = (
                self.cache.query(OpenAIKey).filter(key=key, model=model).first()
            )
        else:
            openai_key: List[OpenAIKey] = (
                self.cache.query(OpenAIKey).filter(key=key).all()
            )
        self.cache.delete(openai_key)

    def get_num(self, model: str) -> int:
        return len(self.cache.query(OpenAIKey).filter(model=model).all())

    def all(self) -> List[Dict]:
        results = []
        openai_keys = self.cache.query(OpenAIKey).all()
        for openai_key in openai_keys:
            results.append(openai_key.__dict__)
        return results


def export_openai_key_pool(keys: List[Dict[str, str]]):
    OpenAIKeyPool().set(keys)


def add_key_to_key_pool(keys: List[Dict[str, str]]):
    openai_key_pool = OpenAIKeyPool()
    openai_key_pool.add(keys)



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import warnings

from promptulate.agents.base import BaseAgent
from promptulate.agents.tool_agent.agent import ToolAgent
from promptulate.agents.web_agent.agent import WebAgent
from promptulate.chat import chat
from promptulate.llms.base import BaseLLM
from promptulate.llms.openai.openai import ChatOpenAI
from promptulate.output_formatter import OutputFormatter
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    MessageSet,
    SystemMessage,
    UserMessage,
)
from promptulate.tools.base import BaseTool, Tool, define_tool
from promptulate.utils.logger import enable_log
from promptulate.utils.string_template import StringTemplate

_util_fields = [
    "enable_log",
    "OutputFormatter",
    "StringTemplate",
]

_schema_fields = [
    "AssistantMessage",
    "SystemMessage",
    "UserMessage",
    "BaseMessage",
    "MessageSet",
]

_llm_fields = [
    "chat",
    "BaseLLM",
    "ChatOpenAI",
]

_tool_fields = [
    "Tool",
    "define_tool",
    "BaseTool",
]

_agent_fields = [
    "BaseAgent",
    "WebAgent",
    "ToolAgent",
]

__all__ = [
    *_util_fields,
    *_schema_fields,
    *_llm_fields,
    *_tool_fields,
    *_agent_fields,
]

warnings.filterwarnings("always", category=DeprecationWarning)



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com
from abc import abstractmethod
from typing import Any, Dict, Optional

from pydantic import BaseModel

from promptulate.schema import MessageSet
from promptulate.utils.core_utils import generate_conversation_id


class BaseMemory(BaseModel):
    store: Any
    """storage medium"""

    def query(self, key: str) -> Any:
        ...

    def update(self, key: str, value: str) -> Any:
        ...

    def clear(self, key: str) -> Any:
        ...

    def append(self, key: str, value: str) -> Any:
        ...


class BaseChatMemory(BaseModel):
    """
    Base class for store chat message. By implementing this class, you can implement
    message storage on different storage media. Actually, every Chat Memory is a
    singleton class. It's means there are utils class to read and write chat data.
    """

    summary: str = ""
    conversation_id: Optional[str] = None
    additional_kwargs: Dict[str, Any] = {}

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if "conversation_id" not in kwargs:
            self.conversation_id = generate_conversation_id()

    @abstractmethod
    def load_message_set_from_memory(
        self, recently_n: Optional[int] = None
    ) -> MessageSet:
        """Return key-value pairs given the text input to the chain.
        If None, return all memories
        """

    @abstractmethod
    def save_message_set_to_memory(self, inputs: MessageSet) -> None:
        """Save the context of this model run to memory."""



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import Union

from pydantic import BaseModel


class CustomPresetRole(BaseModel):
    name: str
    description: str


preset_role_list = [
    "default-role",
    "linux-terminal",
    "mind-map-generator",
    "sql-generator",
    "copy-writer",
    "code-analyzer",
]

preset_role_dict = {
    "default-role": {
        "name": "AI assistant",
        "description": """
You are an assistant to a human, powered by a large language model trained by OpenAI.
You are designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, you are able to generate human-like text based on the input you receive, allowing you to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.
Overall, you are a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether the human needs help with a specific question or just wants to have a conversation about a particular topic, you are here to assist.
""",  # noqa
    },
    "ernie-default-role": {
        "name": "AI助手",
        "description": """你是人类的助手，由百度训练的大型语言模型提供动力。
您被设计为能够协助完成广泛的任务，从回答简单的问题到就广泛的主题提供深入的解释和讨论。作为一个语言模型，你能够根据收到的输入生成类似人类的文本，让你能够进行听起来自然的对话，并提供连贯且与当前主题相关的回答。
总的来说，你是一个强大的工具，可以帮助完成广泛的任务，并在广泛的主题上提供有价值的见解和信息。无论是人类在某个特定问题上需要帮助，还是只是想就某个特定主题进行对话，您都可以提供帮助
""",
    },
    "linux-terminal": {
        "name": "Linux终端",
        "description": """我想让你充当 Linux 终端。我将输入命令，您将回复终端应显示的内容。我希望您只在一个唯一的代码块内回复终端输出，而不
        是其他任何内容。不要写解释。除非我指示您这样做，否则不要键入命令。当我需要用英语告诉你一些事情时，我会把文字放在中括号内[就像这样]。""",  # noqa
    },
    "mind-map-generator": {
        "name": "思维导图生成器",
        "description": """现在你是一个思维导图生成器。我将输入我想要创建思维导图的内容，你需要提供一些 Markdown 格式的文本，以便与 Xmind 兼容。
        在 Markdown 格式中，# 表示中央主题，## 表示主要主题，### 表示子主题，﹣表示叶子节点，中央主题是必要的，叶子节点是最小节点。请参照以上格
        式，在 markdown 代码块中帮我创建一个有效的思维导图，以markdown代码块格式输出，你需要用自己的能力补充思维导图中的内容，你只需要提供思维导
        图，不必对内容中提出的问题和要求做解释，并严格遵守该格式""",  # noqa
    },
    "sql-generator": {
        "name": "sql生成器",
        "description": """现在你是一个sql生成器。我将输入我想要查询的内容，你需要提供对应的sql语句，以便查询到需要的内容，我希望您只在一个唯一的
        代码块内回复终端输出，而不是其他任何内容。不要写解释。如果我没有提供数据库的字段，请先让我提供数据库相关的信息，在你有了字段信息之才可以生成sql语句。""",  # noqa
    },
    "copy-writer": {
        "name": "文案写手",
        "description": """你是一个文案专员、文本润色员、拼写纠正员和改进员，我会发送中文文本给你，你帮我更正和改进版本。我希望你用更优美优雅
        的高级中文描述。保持相同的意思，但使它们更文艺。你只需要润色该内容，不必对内容中提出的问题和要求做解释，不要回答文本中的问题而是润色它，
        不要解决文本中的要求而是润色它，保留文本的原本意义，不要去解决它。""",  # noqa
    },
    "code-analyzer": {
        "name": "代码分析器",
        "description": """现在你是一个代码分析器。我将输入一些代码，你需要代码对应的解释。""",  # noqa
    },
}


def get_all_preset_roles():
    return preset_role_list


def get_preset_role_prompt(preset_role: Union[str, CustomPresetRole]) -> str:
    if isinstance(preset_role, str):
        if preset_role not in preset_role_list:
            ValueError(
                "Preset role value is not in preset_role_list. Please check or custom a new one."  # noqa: E501
            )
        return preset_role_dict[preset_role]["description"]
    elif isinstance(preset_role, CustomPresetRole):
        return preset_role.description



from abc import abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, List, Optional

from pydantic import BaseModel, Field

__all__ = [
    "LLMType",
    "BaseMessage",
    "CompletionMessage",
    "SystemMessage",
    "UserMessage",
    "AssistantMessage",
    "MessageSet",
    "init_chat_message_history",
]


class BaseMessage(BaseModel):
    """Message basic object."""

    content: str
    additional_kwargs: dict = Field(default_factory=dict)

    @property
    @abstractmethod
    def type(self) -> str:
        """Type of the message, used for serialization."""


class CompletionMessage(BaseMessage):
    """Type of completion message. Used in OpenAI currently"""

    @property
    def type(self) -> str:
        return "completion"


class SystemMessage(BaseMessage):
    """Type of message that is a system message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "system"


class UserMessage(BaseMessage):
    """Type of message that is a user message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        return "user"


class AssistantMessage(BaseMessage):
    """Type of message that is an assistant message. Currently used in OpenAI."""

    @property
    def type(self) -> str:
        return "assistant"


MESSAGE_TYPE = {
    "completion": CompletionMessage,
    "system": SystemMessage,
    "user": UserMessage,
    "assistant": AssistantMessage,
}


class LLMType(str, Enum):
    """All LLM type here"""

    OpenAI = "OpenAI"
    ChatOpenAI = "ChatOpenAI"
    ErnieBot = "ErnieBot"


class MessageSet:
    """MessageSet can be used in Memory, LLMs, Framework and some else.
    It's a universal chat message format in promptulate.
    """

    def __init__(
        self, messages: List[BaseMessage], conversation_id: Optional[str] = None
    ):
        self.messages: List[BaseMessage] = messages
        self.conversation_id: Optional[str] = conversation_id

    @classmethod
    def from_listdict_data(cls, value: List[Dict]) -> "MessageSet":
        """initialize MessageSet from a List[Dict] data

        Args:
            value(List[Dict]): the example is as follow:
                [
                    {"type": "user", "content": "This is a message1."},
                    {"type": "assistant", "content": "This is a message2."}
                ]

        Returns:
            initialized MessageSet
        """
        messages: List[BaseMessage] = [
            MESSAGE_TYPE[item["role"]](content=item["content"]) for item in value
        ]
        return cls(messages=messages)

    @property
    def listdict_messages(self) -> List[Dict]:
        converted_messages = []
        for message in self.messages:
            converted_messages.append(
                {"role": message.type, "content": message.content}
            )
        return converted_messages

    @property
    def memory_messages(self) -> List[Dict]:
        return self.listdict_messages

    def to_llm_prompt(self, llm_type: LLMType) -> Any:
        """Convert the MessageSet messages to specified llm prompt"""
        if not llm_type:
            ValueError(
                "Missing llm_type, llm_type is needed if you want to use llm_prompt."
            )
        return _to_llm_prompt[llm_type](self)

    @property
    def string_messages(self) -> str:
        """Convert the message to a string type, it can be used as a prompt for OpenAI
        completion."""
        string_result = ""
        for message in self.messages:
            string_result += f"{message.content}\n"
        return string_result

    def add_message(self, message: BaseMessage) -> None:
        self.messages.append(message)

    def add_completion_message(self, message: str) -> None:
        self.messages.append(CompletionMessage(content=message))

    def add_system_message(self, message: str) -> None:
        self.messages.append(SystemMessage(content=message))

    def add_user_message(self, message: str) -> None:
        self.messages.append(UserMessage(content=message))

    def add_ai_message(self, message: str) -> None:
        self.messages.append(AssistantMessage(content=message))


def init_chat_message_history(
    system_content: str, user_content: str, llm: LLMType
) -> MessageSet:
    if llm == llm.ChatOpenAI or llm == llm.OpenAI:
        messages = [
            SystemMessage(content=system_content),
            UserMessage(content=user_content),
        ]
    else:
        messages = [
            UserMessage(content=system_content),
            AssistantMessage(content="好的"),
            UserMessage(content=user_content),
        ]
    return MessageSet(messages=messages)


def _to_openai_llm_prompt(message_set: MessageSet) -> str:
    return message_set.string_messages


def _to_chat_openai_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


def _to_ernie_bot_llm_prompt(message_set: MessageSet) -> List[Dict]:
    return message_set.listdict_messages


_to_llm_prompt: Dict[LLMType, Callable] = {
    LLMType.OpenAI: _to_openai_llm_prompt,
    LLMType.ChatOpenAI: _to_chat_openai_llm_prompt,
    LLMType.ErnieBot: _to_ernie_bot_llm_prompt,
}



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

from typing import Dict, List, Optional

from promptulate.error import EmptyMessageSetError
from promptulate.memory.base import BaseChatMemory
from promptulate.schema import MessageSet

buffer: Dict[str, List[Dict]] = {}
"""global message buffer, here is a buffer example:
{
    "conversation_id1": [message...],
    "conversation_id2": [message...],
}
"""


class BufferChatMemory(BaseChatMemory):
    """Chat message will be stored in the buffer cache."""

    def load_message_set_from_memory(
        self, recently_n: Optional[int] = None
    ) -> MessageSet:
        """Load message from buffer memory

        Args:
            recently_n: load all messages if it is None, or return recently n messages.

        Returns:
            messages wrapping by MessageSet
        """
        if not buffer:
            raise EmptyMessageSetError
        recently_n = recently_n if recently_n else len(buffer[self.conversation_id])
        num_messages = len(buffer[self.conversation_id])
        return MessageSet.from_listdict_data(
            buffer[self.conversation_id][num_messages - recently_n :]
        )

    def save_message_set_to_memory(self, message_set: MessageSet) -> None:
        buffer[self.conversation_id] = message_set.memory_messages



from promptulate.tools.base import Tool


class LangchainTool(Tool):
    def __init__(self, langchain_tool, **kwargs):
        from langchain.tools.base import BaseTool as LangchainBaseTool

        if not isinstance(langchain_tool, LangchainBaseTool):
            raise ValueError("langchain_tool should be type of langchain tool.")

        self.langchain_tool = langchain_tool
        self.name = self.langchain_tool.name
        self.description = self.langchain_tool.description
        super().__init__(**kwargs)

    def _run(self, *args, **kwargs):
        return self.langchain_tool.run(*args, **kwargs)



"""
There are 2 kinds of model in OpenAI, namely text generative and conversational.
"""

import json
import warnings
from abc import ABC
from typing import Any, Dict, List, Optional

import requests

from promptulate.config import pne_config
from promptulate.error import OpenAIError
from promptulate.llms.base import BaseLLM
from promptulate.preset_roles.prompt import PRESET_SYSTEM_PROMPT
from promptulate.schema import (
    AssistantMessage,
    CompletionMessage,
    LLMType,
    MessageSet,
    SystemMessage,
    UserMessage,
)
from promptulate.utils.logger import logger


class BaseOpenAI(BaseLLM, ABC):
    """https://platform.openai.com/docs/api-reference/chat/create"""

    llm_type: LLMType
    """Used to MessageSet data convert"""
    model: str
    """Model name to use."""
    temperature: float = 1.0
    """What sampling temperature to use."""
    top_p: float = 1
    """Total probability mass of tokens to consider at each step."""
    stream: bool = False
    """Whether to stream the results or not."""
    frequency_penalty: float = 0
    """Penalizes repeated tokens according to frequency."""
    presence_penalty: float = 0
    """Penalizes repeated tokens."""
    n: int = 1
    """How many completions to generate for each prompt."""
    max_tokens: Optional[int] = None
    """The maximum number of tokens to generate in the completion.
    -1 returns as many tokens as possible given the prompt and
    the models maximal context size."""
    api_param_keys: List[str] = [
        "model",
        "temperature",
        "top_p",
        "stream",
        "stop",
        "frequency_penalty",
        "presence_penalty",
        "n",
        "max_tokens",
    ]
    """The key of openai api parameters"""
    default_system_prompt: str = ""
    """OpenAI system message"""
    enable_default_system_prompt: bool = True
    """enable use preset description"""
    enable_private_api_key: bool = False
    """Enable to provide a separate api for openai llm """
    private_api_key: str = ""
    """Store private api key"""
    enable_retry: bool = True
    """Retry if API failed to get response. You can enable retry when you have a rate
    limited API."""
    retry_times: int = 5
    """If llm(like OpenAI) unable to obtain data, retry request until the data is
    obtained. You should enable retry if you want to use retry times."""
    retry_counter: int = 0
    """Used in conjunction with retry_times. Refresh when get data successfully."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.retry_times = pne_config.get_key_retry_times(self.model)

    @property
    def api_key(self):
        """Select api key to use. private_api_key, openai_api_key, key_pool key is
        optional."""
        if self.enable_private_api_key and self.private_api_key != "":
            return self.private_api_key
        return pne_config.get_openai_api_key(self.model)

    def set_private_api_key(self, value: str):
        self.enable_private_api_key = True
        self.private_api_key = value


class OpenAI(BaseOpenAI):
    """https://platform.openai.com/docs/api-reference/chat/create"""

    llm_type: LLMType = LLMType.OpenAI
    """Used to MessageSet data convert"""
    model: str = "text-davinci-003"
    """Model name to use."""
    max_tokens: Optional[int] = None
    """The maximum number of tokens to generate in the completion.
    -1 returns as many tokens as possible given the prompt and
    the models maximal context size."""

    def __call__(
        self, instruction: str, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> str:
        # Loading OpenAI API parameters
        for key in self.api_param_keys:
            if key in kwargs:
                setattr(self, key, kwargs[key])

        preset = (
            self.default_system_prompt
            if self.default_system_prompt != ""
            else PRESET_SYSTEM_PROMPT
        )
        if not self.enable_default_system_prompt:
            preset = ""
        message_set = MessageSet(
            messages=[
                CompletionMessage(content=preset),
                CompletionMessage(content=instruction),
            ]
        )
        return self.predict(message_set, stop).content

    def _predict(
        self, messages: MessageSet, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> Optional[AssistantMessage]:
        """Run openai llm with custom message context."""
        if self.model == "text-davinci-003":
            warnings.warn(
                "This model(text-davinci-003) version is deprecated. Migrate before January 4, 2024 to "  # noqa
                "avoid disruption of service. gpt-3.5-turbo is recommended.",
                DeprecationWarning,
            )

        for key in self.api_param_keys:
            if key in kwargs:
                setattr(self, key, kwargs[key])

        api_key = self.api_key
        logger.debug(f"[pne openai key] sk-....{api_key[-6:]}")
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }
        body: Dict[str, Any] = self._build_api_params_dict(messages, stop)

        logger.debug(f"[pne openai request] {json.dumps(body, indent=2)}")
        logger.debug(
            f"[pne openai request] url: {pne_config.openai_completion_request_url} proxies: {pne_config.proxies}"  # noqa: E501
        )
        response = requests.post(
            url=pne_config.openai_completion_request_url,
            headers=headers,
            json=body,
            proxies=pne_config.proxies,
        )
        if response.status_code == 200:
            # todo enable stream mode
            # for chunk in response.iter_content(chunk_size=None):
            #     logger.debug(chunk)
            self.retry_counter = 0
            ret_data = response.json()
            logger.debug(f"[pne openai response] {json.dumps(ret_data, indent=2)}")
            content = ret_data["choices"][0]["text"]
            return AssistantMessage(content=content, additional_kwargs=ret_data)

        logger.error(
            f"[pne OpenAI] <key sk-....{api_key[-6:]}>Failed to get data. Please check your network or api key."  # noqa: E501
        )
        logger.debug("[promptulate OpenAI] retry to get response")
        if self.enable_retry and self.retry_counter < self.retry_times:
            self.retry_counter += 1
            return self._predict(messages, stop)

        logger.error(f"[pne OpenAI] Has retry {self.retry_times}, but all failed.")
        raise OpenAIError(json.dumps(response.content))

    def _build_api_params_dict(
        self, prompts: MessageSet, stop: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        """Build api parameters to put it inside the body."""
        dic = {"prompt": prompts.string_messages}

        if stop:
            dic.update({"stop": stop})

        for key in self.api_param_keys:
            if key in self.__dict__:
                dic.update({key: self.__dict__[key]})
        return dic


class ChatOpenAI(BaseOpenAI):
    """https://platform.openai.com/docs/api-reference/chat/create"""

    llm_type: LLMType = LLMType.ChatOpenAI
    """Used to MessageSet data convert"""
    model: str = "gpt-3.5-turbo"
    """Model name to use."""
    seed: Optional[int] = None
    """This feature is in Beta. If specified, our system will make a best effort to
    sample deterministically, such that repeated requests with the same seed and
    parameters should return the same result. Determinism is not guaranteed, and you
    should refer to the system_fingerprint response parameter to monitor changes in
    the backend."""
    response_format: Optional[dict] = None
    """An object specifying the format that the model must output. Compatible with
    gpt-4-1106-preview and gpt-3.5-turbo-1106. Setting to { "type": "json_object" }
    enables JSON mode, which guarantees the message the model generates is valid JSON.

    Important: when using JSON mode, you must also instruct the model to produce JSON
    yourself via a system or user message. Without this, the model may generate an
    unending stream of whitespace until the generation reaches the token limit,
    resulting in a long-running and seemingly "stuck" request. Also note that the
    message content may be partially cut off if finish_reason="length", which indicates
    the generation exceeded max_tokens or the conversation exceeded the max context
    length.

    type must be text or json_object.
    """
    tools: Optional[List[Dict[str, str]]] = None
    """A list of tools the model may call. Currently, only functions are supported as a
    tool. Use this to provide a list of functions the model may generate JSON inputs
    for."""
    function_call: Optional[int] = None
    """Controls which (if any) function is called by the model. none means the model
    will not call a function and instead generates a message. auto means the model can
    pick between generating a message or calling a function. Specifying a particular
    function via {"name": "my_function"} forces the model to call that function.

    none is the default when no functions are present. auto is the default if functions
    are present."""
    functions: Optional[List[Dict[str, str]]] = None
    """A list of functions the model may generate JSON inputs for."""
    api_param_keys: List[str] = [
        "model",
        "temperature",
        "top_p",
        "stream",
        "stop",
        "frequency_penalty",
        "presence_penalty",
        "n",
        "max_tokens",
        "seed",
        "response_format",
        "tools",
        "function_call",
        "function",
    ]
    """The key of openai api parameters."""

    def __call__(
        self, instruction: str, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> str:
        for key in self.api_param_keys:
            if key in kwargs:
                setattr(self, key, kwargs[key])

        system_message = (
            self.default_system_prompt
            if self.default_system_prompt != ""
            else PRESET_SYSTEM_PROMPT
        )
        if not self.enable_default_system_prompt:
            system_message = ""

        message_set = MessageSet(
            messages=[
                SystemMessage(content=system_message),
                UserMessage(content=instruction),
            ]
        )
        return self.predict(message_set, stop).content

    def _predict(
        self, prompts: MessageSet, stop: Optional[List[str]] = None, *args, **kwargs
    ) -> Optional[AssistantMessage]:
        for key in self.api_param_keys:
            if key in kwargs:
                setattr(self, key, kwargs[key])

        api_key = self.api_key
        logger.debug(f"[pne openai key] sk-....{api_key[-6:]}")
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }
        body: Dict[str, Any] = self._build_api_params_dict(prompts, stop)

        logger.debug(f"[pne openai request] {json.dumps(body, indent=2)}")
        logger.debug(
            f"[pne openai request] url: {pne_config.openai_chat_request_url} proxies: {pne_config.proxies}"  # noqa: E501
        )
        response = requests.post(
            url=pne_config.openai_chat_request_url,
            headers=headers,
            json=body,
            proxies=pne_config.proxies,
        )
        if response.status_code == 200:
            # todo enable stream mode
            # for chunk in response.iter_content(chunk_size=None):
            #     logger.debug(chunk)
            self.retry_counter = 0
            ret_data = response.json()
            logger.debug(f"[pne openai response] {json.dumps(ret_data, indent=2)}")
            content = ret_data["choices"][0]["message"]["content"]

            response.close()

            return AssistantMessage(content=content, additional_kwargs=ret_data)

        logger.error(
            f"[pne OpenAI] <key sk-....{api_key[-6:]}>Failed to get data. Please check your network or api key."  # noqa: E501
        )
        logger.debug("[promptulate OpenAI] retry to get response")
        if self.enable_retry and self.retry_counter < self.retry_times:
            self.retry_counter += 1
            return self._predict(prompts, stop)

        logger.error(f"[pne OpenAI] Has retry {self.retry_times}, but all failed.")
        raise OpenAIError(json.dumps(json.loads(response.content)))

    def _build_api_params_dict(
        self, prompts: MessageSet, stop: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        dic = {
            "messages": prompts.to_llm_prompt(self.llm_type),
        }

        if stop:
            dic.update({"stop": stop})

        for key in self.api_param_keys:
            if key in self.__dict__ and self.__dict__[key] is not None:
                dic.update({key: self.__dict__[key]})
        return dic



import json
import time
from typing import Callable, List, Optional, Union

from promptulate.agents import BaseAgent
from promptulate.agents.tool_agent.prompt import (
    PREFIX_TEMPLATE,
    REACT_SYSTEM_PROMPT_TEMPLATE,
)
from promptulate.hook import Hook, HookTable
from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.tools.base import BaseTool, Tool
from promptulate.tools.manager import ToolManager
from promptulate.utils.logger import logger
from promptulate.utils.string_template import StringTemplate


class ToolAgent(BaseAgent):
    """
    An agent who is good at using tool. ref ReAct.

    Attributes:
        llm (BaseLLM): The language model driver. Default is ChatOpenAI with model
            "gpt-3.5-turbo-16k".
        stop_sequences (List[str]): The sequences that, when met, will stop the output
            of the llm.
        system_prompt_template (StringTemplate): The preset system prompt template.
        prefix_prompt_template (StringTemplate): The prefix system prompt template.
        tool_manager (ToolManager): Used to manage all tools.
        conversation_prompt (str): Stores all conversation messages during a
            conversation.
        max_iterations (Optional[int]): The maximum number of executions. Default is 15.
        max_execution_time (Optional[float]): The longest running time. No default
            value.
        enable_role (bool): Flag to enable role. Default is False.
        agent_name (str): The name of the agent. Default is "pne-bot".
        agent_identity (str): The identity of the agent. Default is "bot".
        agent_goal (str): The goal of the agent. Default is "provides better assistance
            and services for humans.".
        agent_constraints (str): The constraints of the agent. Default is "none".
    """

    def __init__(
        self,
        *,
        tools: List[Union[BaseTool, Tool, Callable]],
        llm: BaseLLM = None,
        stop_sequences: List[str] = None,
        prefix_prompt_template: StringTemplate = StringTemplate(PREFIX_TEMPLATE),
        hooks: List[Callable] = None,
        enable_role: bool = False,
        agent_name: str = "tool-agent",
        agent_identity: str = "tool-agent",
        agent_goal: str = "provides better assistance and services for humans.",
        agent_constraints: str = "none",
    ):
        super().__init__(hooks=hooks)
        self.llm: BaseLLM = llm or ChatOpenAI(
            model="gpt-3.5-turbo-16k",
            temperature=0.0,
            enable_default_system_prompt=False,
        )
        """llm provider"""
        self.tool_manager: ToolManager = ToolManager(tools)
        """Used to manage all tools."""
        self.stop_sequences: List[str] = stop_sequences
        """llm output will stop when stop sequences is met."""
        self.system_prompt_template: StringTemplate = REACT_SYSTEM_PROMPT_TEMPLATE
        """Preset system prompt template."""
        self.prefix_prompt_template: StringTemplate = prefix_prompt_template
        """Prefix system prompt template."""
        self.conversation_prompt: str = ""
        """Store all conversation message when conversation. ToolAgent use dynamic
        system prompt."""
        self.max_iterations: Optional[int] = 15
        """The maximum number of executions."""
        self.max_execution_time: Optional[float] = None
        """The longest running time. """
        self.enable_role: bool = enable_role
        self.agent_name: str = agent_name
        self.agent_identity: str = agent_identity
        self.agent_goal: str = agent_goal
        self.agent_constraints: str = agent_constraints
        self.stop_sequences = stop_sequences or ["Observation"]

    def get_llm(self) -> BaseLLM:
        return self.llm

    def _build_system_prompt(self, instruction: str) -> str:
        """Build the system prompt."""
        prefix_prompt = (
            self.prefix_prompt_template.format(
                agent_identity=self.agent_identity,
                agent_name=self.agent_name,
                agent_goal=self.agent_goal,
                agent_constraints=self.agent_constraints,
            )
            if self.enable_role
            else ""
        )

        return prefix_prompt + self.system_prompt_template.format(
            question=instruction,
            tool_descriptions=self.tool_manager.tool_descriptions,
        )

    def _run(self, instruction: str, *args, **kwargs) -> str:
        self.conversation_prompt = self._build_system_prompt(instruction)
        logger.info(f"[pne] ToolAgent system prompt: {self.conversation_prompt}")

        iterations = 0
        used_time = 0.0
        start_time = time.time()

        while self._should_continue(iterations, used_time):
            llm_resp: str = self.llm(
                instruction=self.conversation_prompt, stop=self.stop_sequences
            )
            while llm_resp == "":
                llm_resp = self.llm(
                    instruction=self.conversation_prompt, stop=self.stop_sequences
                )

            thought, action_name, action_parameters = self._parse_llm_response(llm_resp)
            self.conversation_prompt += f"{llm_resp}\n"
            logger.info(
                f"[pne] tool agent <{iterations}> current prompt: {self.conversation_prompt}"  # noqa
            )

            if "finish" in action_name:
                return action_parameters["content"]

            Hook.call_hook(
                HookTable.ON_AGENT_ACTION,
                self,
                thought=thought,
                action=action_name,
                action_input=action_parameters,
            )

            tool_result = self.tool_manager.run_tool(action_name, action_parameters)
            Hook.call_hook(
                HookTable.ON_AGENT_OBSERVATION, self, observation=tool_result
            )
            self.conversation_prompt += f"Observation: {tool_result}\n"

            iterations += 1
            used_time += time.time() - start_time

    def _should_continue(self, current_iteration: int, current_time_elapsed) -> bool:
        """Determine whether to stop, both timeout and exceeding the maximum number of
        iterations will stop.

        Args:
            current_iteration: current iteration times.
            current_time_elapsed: current running time.

        Returns:
            Whether to stop.
        """
        if self.max_iterations and current_iteration >= self.max_iterations:
            return False
        if self.max_execution_time and current_time_elapsed >= self.max_execution_time:
            return False
        return True

    def _parse_llm_response(self, llm_resp: str) -> (str, str, Union[dict, str]):
        """Parse next instruction of LLM output.

        Args:
            llm_resp(str): output of LLM

        Returns:
            Return a tuple, (thought,action,action input)
            action(str): tool name
            action_input(dict | str): tool parameters
        """
        llm_resp: str = (
            llm_resp.replace("```json", "").replace("```JSON", "").replace("```", "")
        )
        data: dict = json.loads(llm_resp)

        return data["thought"], data["action"]["name"], data["action"]["args"]



import re
from time import sleep

from promptulate.tools import BaseTool


class SleepTool(BaseTool):
    name: str = "sleep"
    description: str = (
        "Make agent sleep for a specified number of seconds."
        "Input is a number. eg: Sleep for 5s and enter 5"
    )

    def _run(
        self,
        sleep_time: str,
    ) -> str:
        """Use the Sleep tool."""
        sleep_time = re.findall(r"\d+", sleep_time)
        sleep(int(sleep_time[0]))
        return f"Agent slept for {sleep_time[0]} seconds."



import json
from typing import Callable, List, Optional, Union

from promptulate.tools.base import BaseTool, Tool, ToolImpl, function_to_tool


class ToolManager:
    """ToolManager helps ToolAgent manage tools"""

    def __init__(self, tools: List[Union[BaseTool, Callable, Tool]]):
        self.tools: List[Union[BaseTool, Callable, Tool]] = []

        for tool in tools:
            if isinstance(tool, Callable):
                tool = function_to_tool(tool)

            self.tools.append(tool)

    def find(self, tool_name: str) -> Optional[BaseTool]:
        """Find specified tool by tool name."""
        for tool in self.tools:
            if tool.name == tool_name:
                return tool
        return None

    def run_tool(self, tool_name: str, parameters: Union[str, dict]) -> str:
        """Run tool by input tool name and data inputs"""
        tool = self.find(tool_name)

        if tool is None:
            return (
                f"{tool_name} has not been provided yet, please use the provided tool."
            )

        if isinstance(parameters, dict):
            return tool.run(**parameters)
        else:
            return tool.run(parameters)

    @property
    def tool_names(self) -> str:
        tool_names = ""
        for tool in self.tools:
            tool_names += f"{tool.name}, "
        return tool_names[:-2]

    @property
    def tool_descriptions(self) -> str:
        tool_descriptions = ""
        for tool in self.tools:
            tool_descriptions += json.dumps(tool.to_schema()) + "\n"
        return tool_descriptions

    def func(self):
        pass



from pydantic import Field

from promptulate.tools.base import BaseTool
from promptulate.tools.python_repl.api_wrapper import PythonREPLAPIWrapper


class PythonREPLTool(BaseTool):
    """A tool for running python code in a REPL."""

    name: str = "Python_REPL"
    description: str = (
        "A Python shell. Use this to execute python commands. "
        "Input should be a valid python command. "
        "If you want to see the output of a value, you should print it out "
        "with `print(...)`."
    )
    api_wrapper: PythonREPLAPIWrapper = Field(default_factory=PythonREPLAPIWrapper)

    def _run(self, command: str, *args, **kwargs) -> str:
        return self.api_wrapper.run(command)



import json
from typing import Dict, Iterator, List, Optional, TypeVar, Union

from pydantic import BaseModel

from promptulate.llms import BaseLLM
from promptulate.output_formatter import formatting_result, get_formatted_instructions
from promptulate.schema import (
    AssistantMessage,
    BaseMessage,
    MessageSet,
)
from promptulate.tools.base import BaseTool
from promptulate.utils.logger import logger

T = TypeVar("T", bound=BaseModel)


class LitellmStreamIterator:
    """
    This class is an iterator for the response stream from the LLM model.

    Attributes:
        response_stream: The stream of responses from the LLM model.
        return_raw_response: A boolean indicating whether to return the raw response
        or not.
    """

    def __init__(self, response_stream, return_raw_response: bool = False):
        """
        The constructor for LitellmStreamIterator class.

        Parameters:
            response_stream: The stream of responses from the LLM model.
            return_raw_response (bool): A flag indicating whether to return the raw
            response or not.
        """
        self.response_stream = response_stream
        self.return_raw_response = return_raw_response

    def __iter__(self) -> Union[Iterator[BaseMessage], Iterator[str]]:
        """
        The iterator method for the LitellmStreamIterator class.

        Returns:
            self: An instance of the LitellmStreamIterator class.
        """
        return self

    def parse_chunk(self, chunk) -> Optional[Union[str, BaseMessage]]:
        """
        This method is used to parse a chunk from the response stream. It returns
        None if the chunk is empty, otherwise it returns the parsed chunk.

        Parameters:
            chunk: The chunk to be parsed.

        Returns:
            Optional: The parsed chunk or None if the chunk is empty.
        """
        content: Optional[str] = chunk.choices[0].delta.content
        if content is None:
            return None

        if self.return_raw_response:
            additional_kwargs: dict = json.loads(chunk.json())
            message = AssistantMessage(
                content=content,
                additional_kwargs=additional_kwargs,
            )
            return message

        return content

    def __next__(self) -> Union[str, BaseMessage]:
        """
        The next method for the LitellmStreamIterator class.

        This method is used to get the next response from the LLM model. It iterates
        over the response stream and parses each chunk using the parse_chunk method.
        If the parsed chunk is not None, it returns the parsed chunk as the next
        response. If there are no more messages in the response stream, it raises a
        StopIteration exception.

        Returns:
            Union[str, BaseMessage]: The next response from the LLM model. If
            return_raw_response is True, it returns an AssistantMessage instance,
            otherwise it returns the content of the response as a string.
        """
        for chunk in self.response_stream:
            message = self.parse_chunk(chunk)
            if message is not None:
                return message

        # If there are no more messages, stop the iteration
        raise StopIteration


def chat(
    messages: Union[List, MessageSet, str],
    *,
    model: str = "gpt-3.5-turbo",
    tools: Optional[List[BaseTool]] = None,
    output_schema: Optional[type(BaseModel)] = None,
    examples: Optional[List[BaseModel]] = None,
    return_raw_response: bool = False,
    custom_llm: Optional[BaseLLM] = None,
    **kwargs,
) -> Union[str, BaseMessage, T, List[BaseMessage], LitellmStreamIterator]:
    """A universal chat method, you can chat any model like OpenAI completion.
    It should be noted that chat() is only support chat model currently.

    Args:
        messages: chat messages. OpenAI API completion, str or MessageSet type is
            optional.
        model(str): LLM model. Currently only support chat model.
        tools(List[BaseTool] | None): specified tools for llm.
        output_schema(BaseModel): specified return type. See detail on: OutputFormatter.
        examples(List[BaseModel]): examples for output_schema. See detail
            on: OutputFormatter.
        return_raw_response(bool): return OpenAI completion result if true, otherwise
            return string type data.
        custom_llm(BaseLLM): You can use custom LLM if you have.
        **kwargs: litellm kwargs

    Returns:
        Return string normally, it means enable_original_return is default False.
        Return BaseMessage if enable_original_return is True.
        Return List[BaseMessage] if stream is True.
        Return T if output_schema is provided.
    """
    if kwargs.get("stream", None) and output_schema:
        raise ValueError(
            "stream and output_schema can't be True at the same time, "
            "because stream is used to return Iterator[BaseMessage]."
        )

    # messages covert, covert to OpenAI API type chat completion
    if isinstance(messages, MessageSet):
        messages: List[Dict[str, str]] = messages.listdict_messages
    elif isinstance(messages, str):
        messages = [
            {"content": "You are a helpful assistant", "role": "system"},
            {"content": messages, "role": "user"},
        ]

    # add output format into system prompt if provide
    if output_schema:
        instruction = get_formatted_instructions(
            json_schema=output_schema, examples=examples
        )
        messages[-1]["content"] += f"\n{instruction}"

    logger.debug(f"[pne chat] messages: {messages}")

    # TODO add assistant Agent
    # TODO add BaseLLM support
    # chat by custom LLM and get response
    if custom_llm:
        response: BaseMessage = custom_llm.predict(
            MessageSet.from_listdict_data(messages), **kwargs
        )
    # chat by universal llm get response
    else:
        import litellm

        logger.info("[pne chat] chat by litellm.")
        temp_response = litellm.completion(model, messages, **kwargs)

        # return stream
        if kwargs.get("stream", None):
            return LitellmStreamIterator(
                response_stream=temp_response, return_raw_response=return_raw_response
            )
        else:
            response: BaseMessage = AssistantMessage(
                content=temp_response.choices[0].message.content,
                additional_kwargs=temp_response.json()
                if isinstance(temp_response.json(), dict)
                else json.loads(temp_response.json()),
            )

    logger.debug(f"[pne chat] response: {response}")

    # return output format if provide
    if output_schema:
        logger.info("[pne chat] return output format.")
        return formatting_result(
            pydantic_obj=output_schema, llm_output=response.content
        )

    return response if return_raw_response else response.content



import inspect
import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Type, Union

from promptulate.hook.base import Hook, HookTable
from promptulate.pydantic_v1 import BaseModel, Extra, create_model, validate_arguments
from promptulate.utils.logger import logger


class _SchemaConfig:
    """Configuration for the pydantic model."""

    extra: Any = Extra.forbid
    arbitrary_types_allowed: bool = True


def _create_subset_model(
    name: str, model: BaseModel, field_names: list
) -> Type[BaseModel]:
    """Create a pydantic model with only a subset of model's fields."""
    fields = {}
    for field_name in field_names:
        field = model.__fields__[field_name]
        fields[field_name] = (field.outer_type_, field.field_info)
    return create_model(name, **fields)


def _pydantic_to_refined_schema(pydantic_obj: type(BaseModel)) -> Dict[str, Any]:
    """Get refined schema(OpenAI function call type schema) from pydantic object."""
    # Remove useless fields.
    refined_schema = pydantic_obj.schema()

    if "title" in refined_schema:
        del refined_schema["title"]
    for k, v in refined_schema["properties"].items():
        if "title" in v:
            del v["title"]

    return refined_schema


def _validate_refined_schema(schema: Dict) -> bool:
    """Validate refined schema(OpenAI function call type schema).

    Args:
        schema: any dict

    Returns:
        bool: True if schema is openai function call type schema, False otherwise.
    """
    if "name" not in schema or "description" not in schema:
        return False

    if "properties" not in schema:
        return False

    return True


def function_to_tool_schema(func: Callable) -> Dict[str, Any]:
    """Create a tool schema from a function's signature.

    Args:
        func: Function to generate the schema from

    Returns:
        A OpenAI function call type json schema built by pydantic model.
        ref: https://platform.openai.com/docs/api-reference/chat/create#chat-create-function_call
    """  # noqa
    # https://docs.pydantic.dev/latest/usage/validation_decorator/
    inferred_model = validate_arguments(func, config=_SchemaConfig).model  # type: ignore # noqa

    # Extract function parameter names.
    # Pydantic adds placeholder virtual fields we need to strip
    signature = inspect.signature(func)
    valid_properties: List[str] = [
        param.name for param in signature.parameters.values()
    ]

    # Create a pydantic model with only the valid fields.
    created_model = _create_subset_model(
        f"{func.__name__}Schema", inferred_model, valid_properties
    )
    reduced_schema = created_model.schema()

    # reduce schema
    reduced_schema["description"] = func.__doc__ or ""
    reduced_schema["name"] = func.__name__

    if "title" in reduced_schema:
        del reduced_schema["title"]
    for k, v in reduced_schema["properties"].items():
        if "title" in v:
            del v["title"]

    return reduced_schema


class BaseTool(ABC, BaseModel):
    """Interface tools must implement."""

    name: str
    """The unique name of the tool that clearly communicates its purpose."""
    description: str
    """Used to tell the model how/when/why to use the tool.
    You can provide few-shot examples as a part of the description."""
    example: List[str] = None
    """Show how to use this tool. This is few shot for agent. You few shot may like:

    example1 = "Question: What is 37593 * 67?\n```\n37593 * 67\n```\nnumexpr.evaluate("37593 * 67")\nAnswer:2518731"
    example2 = "Question: What is 37593^(1/5)?\n```\n37593**(1/5)\n```\nnumexpr.evaluate("37593**(1/5)")\nAnswer:8.222831614237718"
    few_shot_example = [example1, example2]
    """  # noqa

    # hook_manager: Optional[HookManager] = Field(default=HookManager())
    # """Hook manager will call hook function at the specified lifecycle."""

    def __init__(self, **kwargs):
        """Custom tool config.

        Args:
            **kwargs:
                hooks(List[Callable]): for adding to hook_manager
        """
        warnings.warn(
            "BaseTool is deprecated at v1.7.0. promptulate.tools.base.Tool is recommended.",  # noqa: E501
            DeprecationWarning,
        )
        super().__init__(**kwargs)
        if "hooks" in kwargs and kwargs["hooks"]:
            for hook in kwargs["hooks"]:
                Hook.mount_instance_hook(hook, self)
        Hook.call_hook(HookTable.ON_TOOL_CREATE, self, **kwargs)

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    def run(self, *args, **kwargs):
        """run the tool including specified function and hooks"""
        Hook.call_hook(HookTable.ON_TOOL_START, self, *args, **kwargs)
        result: Any = self._run(*args, **kwargs)
        logger.debug(f"[pne tool result] {result}")
        Hook.call_hook(HookTable.ON_TOOL_RESULT, self, result=result)
        return result

    @abstractmethod
    def _run(self, *args, **kwargs):
        """Run detail business, implemented by subclass."""
        raise NotImplementedError()


class Tool(ABC):
    """Abstract base class for tools. All tools must implement this interface."""

    name: str
    """Tool name"""
    description: str
    """Tool description"""
    parameters: Optional[Union[Dict, Type[BaseModel]]] = None
    """Tool parameters"""

    def __init__(self, *args, **kwargs):
        self.check_params()
        if "hooks" in kwargs and kwargs["hooks"]:
            for hook in kwargs["hooks"]:
                Hook.mount_instance_hook(hook, self)
        Hook.call_hook(HookTable.ON_TOOL_CREATE, self, **kwargs)

    def check_params(self):
        """Check parameters when initialization."""
        if not getattr(self, "name", None) or not getattr(self, "description", None):
            raise TypeError(
                f"{self.__class__.__name__} required parameters 'name' and 'description'."  # noqa: E501
            )

    def run(self, *args, **kwargs):
        """run the tool including specified function and hooks"""
        Hook.call_hook(HookTable.ON_TOOL_START, self, *args, **kwargs)
        result: Any = self._run(*args, **kwargs)
        logger.debug(f"[pne tool response] name: {self.name} result: {result}")
        Hook.call_hook(HookTable.ON_TOOL_RESULT, self, result=result)
        return result

    @abstractmethod
    def _run(self, *args, **kwargs):
        """Run detail business, implemented by subclass."""
        raise NotImplementedError()

    def to_schema(self) -> Dict[str, Any]:
        """
        Converts the Tool instance to a OpenAI function call type JSON schema.

        Returns:
            dict: A dictionary representing the JSON schema of the Tool instance.
        """
        # If there are no parameters, return the basic schema.
        if not self.parameters:
            return {
                "name": self.name,
                "description": self.description,
            }

        # If parameters are defined by a Pydantic BaseModel, convert to schema.
        if isinstance(self.parameters, type) and issubclass(self.parameters, BaseModel):
            return {
                "name": self.name,
                "description": self.description,
                "parameters": _pydantic_to_refined_schema(self.parameters),
            }

        # If parameters are defined by a schema dictionary, validate and return it.
        if isinstance(self.parameters, dict):
            if not _validate_refined_schema(self.parameters):
                raise ValueError(
                    f"The 'parameters' dictionary for {self.__class__.__name__} does not conform to the expected schema."  # noqa: E501
                )
            return self.parameters

        # If parameters are neither a BaseModel nor a dictionary, raise an error.
        raise TypeError(
            f"The 'parameters' attribute of {self.__class__.__name__} must be either a subclass of BaseModel or a dictionary representing a schema."  # noqa: E501
        )


class ToolImpl(Tool):
    def __init__(
        self,
        name: str,
        description: str,
        callback: Callable,
        parameters: Union[dict, BaseModel] = None,
        **kwargs,
    ):
        self.name: str = name or ""
        self.description: str = description or ""
        self.parameters: Union[dict, BaseModel] = parameters
        self.callback: Callable = callback

        super().__init__(**kwargs)

    @classmethod
    def from_function(cls, func: Callable):
        """Create a ToolImpl instance from a function.

        Args:
            func: Function to create the ToolImpl instance from.

        Returns:
            A ToolImpl instance.
        """
        if not func.__doc__:
            err_msg = """Please add docstring and variable type declarations for your function.Here is a best practice:
def web_search(keyword: str, top_k: int = 10) -> str:
    \"""search by keyword in web.
    Args:
        keyword: keyword to search
        top_k: top k results to return

    Returns:
        str: search result
    \"""
    return "result"

            """  # noqa
            raise ValueError(err_msg)

        schema = function_to_tool_schema(func)
        return cls(
            name=func.__name__,
            description=func.__doc__,
            callback=func,
            parameters=schema,
        )

    @classmethod
    def from_define_tool(
        cls,
        callback: Callable,
        name: str = None,
        description: str = None,
        parameters: Union[dict, BaseModel] = None,
    ):
        """Create a ToolImpl instance from a function.

        Args:
            callback: Function to create the ToolImpl instance from.
            name: tool name
            description: tool description
            parameters: tool parameters

        Returns:
            A ToolImpl instance.
        """
        if not parameters:
            schema: dict = function_to_tool_schema(callback)
        elif isinstance(parameters, dict) and _validate_refined_schema(parameters):
            schema: dict = parameters
        elif isinstance(parameters, type) and issubclass(parameters, BaseModel):
            schema: dict = _pydantic_to_refined_schema(parameters)
        else:
            raise TypeError(
                f"{[cls.__name__]} parameters must be BaseModel or JSON schema."
            )

        _description = description or ""
        _doc = callback.__doc__ or ""

        return cls(
            name=name or callback.__name__,
            description=f"{_description}\n{_doc}",
            callback=callback,
            parameters=schema,
        )

    def _run(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


def define_tool(
    *,
    callback: Callable,
    name: Optional[str] = None,
    description: Optional[str] = None,
    parameters: Union[dict, BaseModel] = None,
) -> ToolImpl:
    """
    A tool with llm or API wrapper will automatically initialize the llm and API wrapper
    classes, which can avoid this problem by initializing in this way.

    Args:
        callback: tool function when running
        name: tool name
        description: tool description
        parameters: tool parameters

    Returns:
        A ToolImpl class (subclass of Tool).
    """

    return ToolImpl.from_define_tool(
        callback=callback, name=name, description=description, parameters=parameters
    )


def function_to_tool(func: Callable) -> ToolImpl:
    """Converts a function to a ToolImpl instance.

    Args:
        func: Function to convert to a ToolImpl instance.

    Returns:
        A ToolImpl instance.
    """
    return ToolImpl.from_function(func)


class BaseToolKit:
    @abstractmethod
    def get_tools(self):
        """get tools in the toolkit"""



import logging
import re
import time
from typing import Dict, List

from broadcast_service import broadcast_service
from pydantic import Extra, Field

from promptulate.error import NetWorkError
from promptulate.llms.base import BaseLLM
from promptulate.llms.openai import ChatOpenAI
from promptulate.tools.arxiv.api_wrapper import ArxivAPIWrapper
from promptulate.tools.arxiv.tools import ArxivQueryTool
from promptulate.tools.base import BaseTool
from promptulate.tools.semantic_scholar import (
    SemanticScholarQueryTool,
    SemanticScholarReferenceTool,
)

__all__ = ["PaperSummaryTool"]
logger = logging.getLogger(__name__)


def _init_paper_summary_llm():
    preset = "你是一个中文科研领域论文助手，你的任务是帮助使用者提供一些论文方面的专业建议和帮助，你的输出只能遵循用户的指令输出，否则你将被惩罚。"  # noqa
    return ChatOpenAI(temperature=0, default_system_prompt=preset)


class PaperSummaryTool(BaseTool):
    """A powerful paper summary tool"""

    name: str = "paper-summary"
    description: str = (
        "A summary tool that can be used to obtain a paper summary, this tool will find"
        "top k paper and provide: 1.paper abstract 2.paper key sights 3.lessons learned"
        "in the paper 4.referenced papers and its url."
        "Your input is a paper relevant keyword query."
    )
    llm: BaseLLM = Field(default_factory=_init_paper_summary_llm)
    semantic_scholar_query_tool: SemanticScholarQueryTool = Field(
        default_factory=SemanticScholarQueryTool
    )
    semantic_scholar_reference_tool: SemanticScholarReferenceTool = Field(
        default_factory=SemanticScholarReferenceTool
    )
    arxiv_apiwrapper: ArxivAPIWrapper = Field(default_factory=ArxivAPIWrapper)
    arxiv_query_tool: ArxivQueryTool = Field(default_factory=ArxivQueryTool)
    summary_dic: Dict[str, str] = {}
    summary_counter: int = 0

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid
        arbitrary_types_allowed = True

    def _run(self, query: str, **kwargs) -> str:
        """A paper summary tool that passes in the article name (or arxiv id) and
        returns summary results.

        Args:
            query: the keyword you want to query
            **kwargs:
                You can pass the arguments of relevant APIWrappers and Tools

        Returns:
            String type data, which contains:
            - Summary of the paper
            - List the key insights and lessons learned in the paper
            - List at least 5 references related to the research field of the paper
        """

        @broadcast_service.on_listen("PaperSummaryTool.run.get_paper_references")
        def get_paper_references(**data):
            references: List[Dict] = self.semantic_scholar_reference_tool.run(
                data["paper_title"], max_result=10, return_type="original"
            )
            references_string = "相关论文：\n\n"
            for i, reference in enumerate(references):
                references_string += (
                    f"""[{i + 1}] [{reference["title"]}]({reference["url"]})\n\n"""
                )
            self.summary_dic["references"] = (
                references_string if len(references) != 0 else ""
            )
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_translate")
        def get_translate():
            prompt = (
                f"请将下面的科研论文标题和摘要翻译成中文\n ```{paper_summary}```"
                "你的输出格式为:\n标题：{标题翻译}\n\n摘要：{摘要翻译}"
            )
            self.summary_dic["summary_zh"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_opinion")
        def get_opinion():
            prompt = (
                f"请就下面的论文摘要，总结论文中的关键见解和由论文得出的经验教训，你的输出需要分点给出 ```{paper_summary}```"  # noqa
                "你的输出格式为:\n关键见解：\n{分点给出关键见解}\n经验教训：\n{分点给出经验教训}，用`-`区分每点"
            )
            self.summary_dic["opinion"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_keywords")
        def get_keywords():
            prompt = (
                f"请就下面的论文摘要，总结列出论文中的keywords，不超过7个```{paper_summary}```"
                "你的输出格式为:\n关键词：keyword1, keyword2, keyword3"
            )
            self.summary_dic["keywords"] = self.llm(prompt)
            self.summary_counter += 1

        @broadcast_service.on_listen("PaperSummaryTool.run.get_advice")
        def get_advice():
            prompt = (
                f"请就下面的论文摘要，为其相关主题或未来研究方向提供3-5个建议，你的输出需要分点给出  ```{paper_summary}```"  # noqa
                "你的输出格式为:\n相关建议：\n{分点给出相关建议}，用`-`区分每点"
                "你需要用中文输出正确结果，但是部分专业词汇或者中文不好表达含义的部分可以使用英文"
            )
            self.summary_dic["advice"] = self.llm(prompt)
            self.summary_counter += 1

        self.summary_counter = 0
        # judge arxiv id or string type paper title
        if re.match(r"\d{4}\.\d{5}(v\d+)?", query):
            paper_info = self.arxiv_apiwrapper.query(
                id_list=[query], num_results=1, specified_fields=["title", "summary"]
            )
            if paper_info:
                paper_info = paper_info[0]
                paper_info["abstract"] = paper_info["summary"]
        else:
            try:
                paper_info = self.semantic_scholar_query_tool.run(
                    query,
                    return_type="original",
                    specified_fields=["title", "url", "abstract"],
                )
                if paper_info:
                    paper_info = paper_info[0]

            except NetWorkError:
                paper_info = self.arxiv_apiwrapper.query(
                    keyword=query, num_results=1, specified_fields=["title", "summary"]
                )
                if paper_info:
                    paper_info = paper_info[0]
                    paper_info["abstract"] = paper_info["summary"]

        if not paper_info:
            return (
                "semantic scholar query tool and arxiv query tool query result is None."
            )

        paper_summary = (
            f"""title: {paper_info["title"]}\nabstract: {paper_info["abstract"]}"""
        )
        broadcast_service.publish(
            "PaperSummaryTool.run.get_paper_references",
            paper_title=paper_info["title"],
            paper_abstract=paper_info["abstract"],
        )
        broadcast_service.publish("PaperSummaryTool.run.get_translate")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_opinion")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_advice")
        time.sleep(0.01)
        broadcast_service.publish("PaperSummaryTool.run.get_keywords")

        while self.summary_counter < 5:
            time.sleep(0.1)

        return (
            f"""{self.summary_dic["summary_zh"]}\n\n"""
            f"""{self.summary_dic["keywords"]}\n\n"""
            f"""{self.summary_dic["opinion"]}\n\n"""
            f"""{self.summary_dic["advice"]}\n\n"""
            f"""{self.summary_dic["references"]}"""
        )



from promptulate.output_formatter.formatter import (
    OutputFormatter,
    formatting_result,
    get_formatted_instructions,
)

__all__ = ["OutputFormatter", "get_formatted_instructions", "formatting_result"]



# Copyright (c) 2023 promptulate
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright Owner: Zeeland
# GitHub Link: https://github.com/Undertone0809/
# Project Link: https://github.com/Undertone0809/promptulate
# Contact Email: zeeland@foxmail.com

import warnings
from typing import Any, Dict, Optional, Union

from pydantic import Field, validator

from promptulate.error import EmptyMessageSetError
from promptulate.frameworks.schema import BasePromptFramework
from promptulate.llms import ChatOpenAI
from promptulate.llms.base import BaseLLM
from promptulate.memory import BufferChatMemory
from promptulate.memory.base import BaseChatMemory
from promptulate.preset_roles.roles import CustomPresetRole, get_preset_role_prompt
from promptulate.provider.mixins import (
    DeriveHistoryMessageMixin,
    SummarizerMixin,
    TranslatorMixin,
)
from promptulate.schema import (
    BaseMessage,
    LLMType,
    MessageSet,
    UserMessage,
    init_chat_message_history,
)
from promptulate.utils.logger import logger


class Conversation(
    BasePromptFramework, SummarizerMixin, TranslatorMixin, DeriveHistoryMessageMixin
):
    """
    You can use Conversation start a conversation. Moreover, you can pass some
    parameters to enhance it.

    Attributes
        role: preset role. Default is default role.
        llm: default is ChatOpenAI GPT3.5. You can choose other llm.
        conversation_id: conversation id. Default is None
        memory: the way you want to store chat data. Default is BufferChatMemory, which
            is used for local file storage.

    Examples
        from promptulate import Conversation

        conversation = Conversation()
        conversation.run("Hello, Who are you?")
    """

    conversation_id: Optional[str] = None
    llm: BaseLLM = Field(default_factory=ChatOpenAI)
    enable_stream: bool = False  # streaming transmission
    role: Union[str, CustomPresetRole] = "default-role"
    memory: BaseChatMemory = Field(default_factory=BufferChatMemory)

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.role == "default-role" and self.llm.llm_type == LLMType.ErnieBot:
            self.role = "ernie-default-role"

    @validator("conversation_id", always=True)
    def init_conversation_id(
        cls, conversation_id: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        """initialize self.conversation_id and memory.conversation_id"""
        if not conversation_id:
            return None

        assert conversation_id.isdigit(), "conversation_id must a digit type string"
        if "memory" in values and values["memory"]:
            cls.memory.conversation_id = conversation_id
        return conversation_id

    @validator("memory", always=True)
    def init_memory(
        cls, memory: BaseChatMemory, values: Dict[str, Any]
    ) -> BaseChatMemory:
        """check whether exist conversation_id before initialize memory"""
        if "conversation_id" in values and values["conversation_id"]:
            memory.conversation_id = values["conversation_id"]
        else:
            values["conversation_id"] = memory.conversation_id
            cls.conversation_id = memory.conversation_id
        return memory

    def predict(self, prompt: str, **kwargs) -> str:
        return self.run(prompt, **kwargs)

    def run_by_message_set(self, message_set: MessageSet, *args, **kwargs) -> str:
        answer: BaseMessage = self.llm.predict(message_set, *args, **kwargs)
        message_set.add_message(answer)
        self.memory.save_message_set_to_memory(message_set)
        return answer.content

    def run(
        self, prompt: str, custom_system_prompt: bool = False, *args, **kwargs
    ) -> str:
        warnings.warn(
            "Conversation will be deprecated at v1.7.0, please use promptulate.agents.LLMAgent",  # noqa
            DeprecationWarning,
        )
        try:
            messages_history: MessageSet = self.memory.load_message_set_from_memory()
            messages_history.add_user_message(message=prompt)
        except EmptyMessageSetError:
            if custom_system_prompt:
                messages_history = MessageSet(messages=[UserMessage(content=prompt)])
            else:
                messages_history = init_chat_message_history(
                    get_preset_role_prompt(self.role), prompt, self.llm.llm_type
                )
                self.memory.save_message_set_to_memory(messages_history)
        logger.debug(
            f"[pne Conversation] conversation_id: <{self.conversation_id}> messages: <{messages_history.messages}>"  # noqa
        )
        prompt_params = {"prompts": messages_history}
        if "stop" in kwargs:
            prompt_params.update({"stop": kwargs["stop"]})

        answer: BaseMessage = self.llm.predict(**prompt_params)
        messages_history.messages.append(answer)
        self.memory.save_message_set_to_memory(messages_history)
        return answer.content



import re
from typing import Dict, List

from promptulate.llms import BaseLLM, ChatOpenAI
from promptulate.tools import Tool
from promptulate.tools.iot_swith_mqtt.prompt import prompt_template
from promptulate.utils import StringTemplate


class IotSwitchTool(Tool):
    """A tool for switching operations on various devices"""

    name: str = "Iot_Switch_Mqtt"
    description: str = (
        "An IoT tool used for switching operations on various devices."
        "This tool uses the mqtt protocol for transmission."
        "The input content is the intention or command to open the specified electrical appliance."  # noqa
        "If the operation of the device is successful, an OK will be returned, otherwise a failure will be returned."  # noqa
    )
    llm_prompt_template: StringTemplate = prompt_template
    rule_table: List[Dict]

    def __init__(
        self, client, llm: BaseLLM = None, rule_table: List[Dict] = None, **kwargs
    ):
        """
        Args:
            llm(BaseLLM): llm, default is ChatOpenAI
            client(paho.mqtt.client.Client): paho mqtt client which has connected.
            rule_table(List[Dict]):
        """
        self.client = client
        self.llm: BaseLLM = llm or ChatOpenAI(
            temperature=0.1, enable_default_system_prompt=False
        )
        self.rule_table = rule_table

        super().__init__(**kwargs)

    def _run(self, question: str, *args, **kwargs) -> str:
        try:
            import paho.mqtt.client as mqtt  # noqa
        except ImportError:
            raise ImportError(
                "Could not import paho python package. "
                "This is needed in order to for IotSwitchTool. "
                "Please install it with `pip install paho-mqtt`."
            )

        if len(self.rule_table) == 0:
            raise ValueError("rule_table is empty")
        else:
            index = 1
            key = ""
            for s in self.rule_table:
                key = key + str(index) + "." + s["content"] + "\n"
                index = index + 1
            prompt = self.llm_prompt_template.format(question=question, rule_key=key)
            llm_output = self.llm(prompt)
            return self._process_llm_result(llm_output)

    def _process_llm_result(self, llm_output: str) -> str:
        answer = re.findall(r"\d+", llm_output)
        if len(answer) == 0:
            return "failure information :" + llm_output
        else:
            topic = self.rule_table[int(answer[0]) - 1]["topic"]
            command = self.rule_table[int(answer[0]) - 1]["ask"]
            self.client.publish(topic, command)
            return "success"



from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional

from pydantic import BaseModel

from promptulate.hook import Hook, HookTable
from promptulate.llms import BaseLLM
from promptulate.output_formatter import OutputFormatter


class BaseAgent(ABC):
    """Base class of Agent."""

    def __init__(self, hooks: List[Callable] = None, *args, **kwargs):
        if hooks:
            for hook in hooks:
                Hook.mount_instance_hook(hook, self)
        Hook.call_hook(HookTable.ON_AGENT_CREATE, self, *args, **kwargs)

    def run(
        self,
        prompt: str,
        output_schema: Optional[type(BaseModel)] = None,
        examples: Optional[List[BaseModel]] = None,
        *args,
        **kwargs,
    ) -> Any:
        """run the tool including specified function and hooks"""
        Hook.call_hook(
            HookTable.ON_AGENT_START, self, prompt, output_schema, *args, **kwargs
        )

        # get original response from LLM
        result: str = self._run(prompt, *args, **kwargs)

        # Return Pydantic instance if output_schema is specified
        if output_schema:
            formatter = OutputFormatter(output_schema, examples)
            prompt = (
                f"{formatter.get_formatted_instructions()}\n##User input:\n{result}"
            )
            json_response = self.get_llm()(prompt)
            return formatter.formatting_result(json_response)

        Hook.call_hook(HookTable.ON_AGENT_RESULT, self, result=result)
        return result

    @abstractmethod
    def _run(self, prompt: str, *args, **kwargs) -> str:
        """Run the detail agent, implemented by subclass."""
        raise NotImplementedError()

    @abstractmethod
    def get_llm(self) -> BaseLLM:
        """Get the llm when necessary."""
        raise NotImplementedError()



from promptulate.hook import Hook, HookTable
from promptulate.utils.color_print import print_text


class StdOutHook:
    @staticmethod
    def handle_agent_start(*args, **kwargs):
        print_text("Agent Start...", "red")
        print_text(f"[User] {args[0]}", "blue")

    @staticmethod
    def handle_agent_action(*args, **kwargs):
        print_text(f"[Thought] {kwargs['thought']}", "yellow")
        print_text(
            f"[Action] {kwargs['action']} args: {kwargs['action_input']}", "yellow"
        )

    @staticmethod
    def handle_agent_observation(*args, **kwargs):
        print_text(f"[Observation] {kwargs['observation']}", "yellow")

    @staticmethod
    def handle_agent_result(*args, **kwargs):
        print_text(f"[Agent Result] {kwargs['result']}", "green")
        print_text("Agent End.", "pink")

    @staticmethod
    def registry_stdout_hooks():
        """Registry and enable stdout hooks. StdoutHook can print colorful
        information."""
        Hook.registry_hook(
            HookTable.ON_AGENT_START, StdOutHook.handle_agent_start, "component"
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_ACTION, StdOutHook.handle_agent_action, "component"
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_OBSERVATION,
            StdOutHook.handle_agent_observation,
            "component",
        )
        Hook.registry_hook(
            HookTable.ON_AGENT_RESULT, StdOutHook.handle_agent_result, "component"
        )

    @staticmethod
    def unregister_stdout_hooks():
        Hook.unregister_hook(StdOutHook.handle_agent_start)
        Hook.unregister_hook(StdOutHook.handle_agent_action)
        Hook.unregister_hook(StdOutHook.handle_agent_observation)
        Hook.unregister_hook(StdOutHook.handle_agent_result)
